# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-


def smooth(signal, owidth, edge_truncate=False):
    """Replicates the IDL ``SMOOTH()`` function.

    Parameters
    ----------
    signal : array-like
        The array to be smoothed.
    owidth : :class:`int` or array-like
        Width of the smoothing window.  Can be a scalar or an array with
        length equal to the number of dimensions of `signal`.
    edge_truncate : :class:`bool`, optional
        Set `edge_truncate` to ``True`` to apply smoothing to all points.
        Points near the edge are normally excluded from smoothing.

    Returns
    -------
    array-like
        A smoothed array with the same dimesions and type as `signal`.

    References
    ----------
    https://www.nv5geospatialsoftware.com/docs/smooth.html

    Examples
    --------
    """
    if owidth % 2 == 0:
        width = owidth + 1
    else:
        width = owidth
    if width < 3:
        return signal
    n = signal.size
    istart = int((width-1)/2)
    iend = n - int((width+1)/2)
    w2 = int(width/2)
    s = signal.copy()
    for i in range(n):
        if i < istart:
            if edge_truncate:
                s[i] = (signal[0:istart+i+1].sum() +
                        (istart-i)*signal[0])/float(width)
        elif i > iend:
            if edge_truncate:
                s[i] = (signal[i-istart:n].sum() +
                        (i-iend)*signal[n-1])/float(width)
        else:
            s[i] = signal[i-w2:i+w2+1].sum()/float(width)
    return s



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.misc.
"""
import os
import pytest
import numpy as np
import tempfile
from .. import PydlutilsException
from ..misc import djs_laxisgen, djs_laxisnum, hogg_iau_name, struct_print


def test_djs_laxisgen():
    #
    # 1d
    #
    assert (np.arange(4, dtype='i4') == djs_laxisgen((4,))).all()
    #
    # 2d
    #
    l = np.array([[0, 0, 0, 0],
                  [1, 1, 1, 1],
                  [2, 2, 2, 2],
                  [3, 3, 3, 3]], dtype='i4')
    assert (l == djs_laxisgen((4, 4))).all()
    assert (l.T == djs_laxisgen((4, 4), iaxis=1)).all()
    with pytest.raises(ValueError):
        _ = djs_laxisgen((4, 4), iaxis=2)
    #
    # 3d
    #
    l = np.zeros((4, 4, 4), dtype='i4')
    l[1, :, :] = 1
    l[2, :, :] = 2
    l[3, :, :] = 3
    assert (l == djs_laxisgen((4, 4, 4))).all()
    assert (l.swapaxes(0, 1) == djs_laxisgen((4, 4, 4), iaxis=1)).all()
    assert (l.swapaxes(0, 2) == djs_laxisgen((4, 4, 4), iaxis=2)).all()
    with pytest.raises(ValueError):
        _ = djs_laxisgen((4, 4, 4), iaxis=3)
    #
    # More d
    #
    with pytest.raises(ValueError):
        _ = djs_laxisgen((4, 4, 4, 4))


def test_djs_laxisnum():
    #
    # 1d
    #
    assert (np.zeros((4,), dtype='i4') == djs_laxisnum((4,))).all()
    #
    # 2d
    #
    l = np.array([[0, 0, 0, 0],
                  [1, 1, 1, 1],
                  [2, 2, 2, 2],
                  [3, 3, 3, 3]], dtype='i4')
    assert (l == djs_laxisnum((4, 4))).all()
    assert (l.T == djs_laxisnum((4, 4), iaxis=1)).all()
    with pytest.raises(ValueError):
        _ = djs_laxisnum((4, 4), iaxis=2)
    #
    # 3d
    #
    l = np.zeros((4, 4, 4), dtype='i4')
    l[1, :, :] = 1
    l[2, :, :] = 2
    l[3, :, :] = 3
    assert (l == djs_laxisnum((4, 4, 4))).all()
    assert (l.swapaxes(0, 1) == djs_laxisnum((4, 4, 4), iaxis=1)).all()
    assert (l.swapaxes(0, 2) == djs_laxisnum((4, 4, 4), iaxis=2)).all()
    with pytest.raises(ValueError):
        _ = djs_laxisnum((4, 4, 4), iaxis=3)
    #
    # More d
    #
    with pytest.raises(ValueError):
        _ = djs_laxisnum((4, 4, 4, 4))


def test_hogg_iau_name():
    assert (hogg_iau_name(354.120375, -0.544777778) ==
            'SDSS J233628.89-003241.2')
    assert (hogg_iau_name(354.120375, -0.544777778, prefix='2MASS') ==
            '2MASS J233628.89-003241.2')
    assert (hogg_iau_name(354.120375, -0.544777778, prefix='') ==
            'J233628.89-003241.2')
    assert (hogg_iau_name(354.120375, -0.544777778, precision=0) ==
            'SDSS J233628.8-003241')
    assert (hogg_iau_name(354.120375, -0.544777778, precision=2) ==
            'SDSS J233628.890-003241.20')
    ra = np.array([354.120375, 7.89439, 36.31915, 110.44730])
    dec = np.array([-0.544777778, -0.35157, 0.47505, 39.35352])
    names = hogg_iau_name(ra, dec)
    assert tuple(names) == ('SDSS J233628.89-003241.2',
                            'SDSS J003134.65-002105.6',
                            'SDSS J022516.59+002830.1',
                            'SDSS J072147.35+392112.6')


def test_struct_print():
    slist = np.zeros((5,), dtype=[('a', 'c16'), ('b', bool)])
    with pytest.raises(PydlutilsException):
        lines, css = struct_print(slist, silent=True)
    slist = np.array([(1, 2.34, 'five'),
                      (2, 3.456, 'seven'),
                      (3, -4.5678, 'nine')],
                     dtype=[('a', 'i4'), ('bb', 'f4'), ('ccc', 'S5')])
    lines, css = struct_print(slist, silent=True)
    assert lines[0] == 'a bb           ccc  '
    assert lines[1] == '- ------------ -----'
    assert lines[2] == '1         2.34 five '
    assert lines[3] == '2        3.456 seven'
    assert lines[4] == '3      -4.5678 nine '
    assert len(css) == 0
    lines, css = struct_print(slist, silent=True, alias={'ccc': 'c'})
    assert lines[0] == 'a bb           c    '
    assert lines[1] == '- ------------ -----'
    assert lines[2] == '1         2.34 five '
    assert lines[3] == '2        3.456 seven'
    assert lines[4] == '3      -4.5678 nine '
    assert len(css) == 0
    lines, css = struct_print(slist, silent=True,
                              formatcodes={'a': '{0:02d}'})
    assert lines[0] == 'a  bb           ccc  '
    assert lines[1] == '-- ------------ -----'
    assert lines[2] == '01         2.34 five '
    assert lines[3] == '02        3.456 seven'
    assert lines[4] == '03      -4.5678 nine '
    assert len(css) == 0
    lines, css = struct_print(slist, silent=True, fdigit=3)
    assert lines[0] == 'a bb         ccc  '
    assert lines[1] == '- ---------- -----'
    assert lines[2] == '1       2.34 five '
    assert lines[3] == '2       3.46 seven'
    assert lines[4] == '3      -4.57 nine '
    assert len(css) == 0
    lines, css = struct_print(slist, silent=True, html=True)
    assert lines[0] == '<table>'
    assert lines[1] == '<tr><th>a</th><th>bb</th><th>ccc</th></tr>'
    assert lines[2] == '<tr><td>1</td><td>        2.34</td><td>five </td></tr>'
    assert lines[3] == '<tr><td>2</td><td>       3.456</td><td>seven</td></tr>'
    assert lines[4] == '<tr><td>3</td><td>     -4.5678</td><td>nine </td></tr>'
    assert lines[5] == '</table>'
    assert css[0] == '<style type="text/css">'
    assert css[1] == 'table {'
    assert css[2] == '    border-collapse: collapse;'
    assert css[3] == '}'
    assert css[4] == 'th {'
    assert css[5] == '    padding: 2px;'
    assert css[6] == '    text-align: right;'
    assert css[7] == '    border: 1px solid black;'
    assert css[8] == '    font-weight: bold;'
    assert css[9] == '}'
    assert css[10] == 'td {'
    assert css[11] == '    padding: 2px;'
    assert css[12] == '    text-align: right;'
    assert css[13] == '    border: 1px solid black;'
    assert css[14] == '}'
    assert css[15] == '</style>'
    slist = np.array([(1, 2.34, 'five'),
                      (2, 3.456, 'seven'),
                      (3, -4.5678, 'nine')],
                     dtype=[('a', 'i4'), ('bb', 'f8'), ('ccc', 'S5')])
    lines, css = struct_print(slist, silent=True, ddigit=3)
    assert lines[0] == 'a bb         ccc  '
    assert lines[1] == '- ---------- -----'
    assert lines[2] == '1       2.34 five '
    assert lines[3] == '2       3.46 seven'
    assert lines[4] == '3      -4.57 nine '
    assert len(css) == 0
    with tempfile.NamedTemporaryFile(delete=False) as spf1:
        spf1_name = spf1.name
        lines, css = struct_print(slist, silent=True,
                                  filename=spf1_name)
    with open(spf1_name, 'rb') as f:
        data = f.read().decode('utf-8')
    assert "\n".join(lines)+"\n" == data
    os.remove(spf1_name)
    with tempfile.TemporaryFile() as spf2:
        lines, css = struct_print(slist, silent=True, filename=spf2)
        spf2.seek(0)
        data = spf2.read().decode('utf-8')
    assert "\n".join(lines)+"\n" == data



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlspec2d.spec1d.
"""
import os
import pytest
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from .. import Pydlspec2dException
from ..spec1d import (HMF, log, spec_append, spec_path, template_metadata,
                      wavevector)


@pytest.fixture
def sdss_env(request):
    """Set up spectroscopic pipeline environment variables.
    """
    m = request.getfixturevalue("monkeypatch")
    e = {'BOSS_SPECTRO_REDUX': '/boss/spectro/redux',
         'SPECTRO_REDUX': '/sdss/spectro/redux',
         'RUN2D': 'v1_2_3',
         'RUN1D': 'v1_2_3'}
    for k in e:
        m.setenv(k, e[k])
    return m


def test_findspec():
    """This is just a placeholder for now.
    """
    # slist = findspec(infile='file.in', sdss=True)
    assert True


def test_hmf_init():
    """Test initialization of HMF object
    """
    spec = np.random.random((20, 100))
    invvar = np.random.random((20, 100))
    hmf = HMF(spec, invvar)
    assert hmf.K == 4
    assert log.level == 20  # INFO
    hmf = HMF(spec, invvar, K=6, verbose=True)
    assert hmf.K == 6
    assert log.level == 10  # DEBUG


def test_spec_append():
    spec1 = np.array([[1, 1, 1, 1, 1],
                      [1, 1, 1, 1, 1]])
    spec2 = np.array([[2, 2, 2, 2, 2],
                      [2, 2, 2, 2, 2]])
    s = spec_append(spec1, spec2)
    assert (s == np.array([[1, 1, 1, 1, 1],
                           [1, 1, 1, 1, 1],
                           [2, 2, 2, 2, 2],
                           [2, 2, 2, 2, 2]])).all()
    spec2 = np.array([[2, 2, 2, 2],
                      [2, 2, 2, 2]])
    s = spec_append(spec1, spec2)
    assert (s == np.array([[1, 1, 1, 1, 1],
                           [1, 1, 1, 1, 1],
                           [2, 2, 2, 2, 0],
                           [2, 2, 2, 2, 0]])).all()
    s = spec_append(spec1, spec2, 1)
    assert (s == np.array([[1, 1, 1, 1, 1],
                           [1, 1, 1, 1, 1],
                           [0, 2, 2, 2, 2],
                           [0, 2, 2, 2, 2]])).all()
    spec1 = np.array([[1, 1, 1],
                      [1, 1, 1]])
    spec2 = np.array([[2, 2, 2, 2, 2],
                      [2, 2, 2, 2, 2]])
    s = spec_append(spec1, spec2, -2)
    assert (s == np.array([[0, 0, 1, 1, 1],
                           [0, 0, 1, 1, 1],
                           [2, 2, 2, 2, 2],
                           [2, 2, 2, 2, 2]])).all()


def test_spec_path(sdss_env):
    bsr = os.environ['BOSS_SPECTRO_REDUX']
    run2d = os.environ['RUN2D']
    p = spec_path(123)
    assert p[0] == os.path.join(bsr, run2d, '0123')
    p = spec_path(1234)
    assert p[0] == os.path.join(bsr, run2d, '1234')
    p = spec_path(1234, topdir=bsr, run2d=run2d)
    assert p[0] == os.path.join(bsr, run2d, '1234')
    p = spec_path(np.array([1234, 5678]), topdir=bsr, run2d=run2d)
    assert p[0] == os.path.join(bsr, run2d, '1234')
    assert p[1] == os.path.join(bsr, run2d, '5678')
    p = spec_path(1234, path=bsr)
    assert p[0] == bsr


def test_template_metadata():
    with pytest.raises(Pydlspec2dException):
        slist, metadata = template_metadata('/no/such/file.par')
    inputfile = get_pkg_data_filename('t/test_template_metadata.par')
    slist, metadata = template_metadata(inputfile)
    assert metadata['object'] == 'gal'
    assert not metadata['nonnegative']


def test_wavevector():
    l = wavevector(3, 4, binsz=0.1)
    ll = np.array([3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4.0])
    assert np.allclose(l, ll)
    l = wavevector(3, 4, wavemin=3, binsz=0.1)
    ll = np.array([3.0, 3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7, 3.8, 3.9, 4.0])
    assert np.allclose(l, ll)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.image.
"""
import pytest
import numpy as np
from ..image import djs_maskinterp1, djs_maskinterp


def test_djs_maskinterp1():
    y = np.array([0.0, 1.0, 2.0, 3.0, 4.0], dtype=np.float64)
    # Test all good
    yi = djs_maskinterp1(y, np.zeros(y.shape, dtype=np.int32))
    assert (yi == y).all()
    # Test all bad
    yi = djs_maskinterp1(y, np.ones(y.shape, dtype=np.int32))
    assert (yi == y).all()
    # Test one good value
    yt = np.zeros(y.shape, dtype=y.dtype) + y[0]
    yi = djs_maskinterp1(y, np.arange(len(y)))
    assert (yi == yt).all()
    # Test two bad values
    yi = djs_maskinterp1(y, np.array([0, 1, 0, 1, 0]))
    assert np.allclose(y, yi)
    # Because of the default behavior of np.interp(), the 'const'
    # keyword has no actual effect, because it performs the
    # same operation.
    yt = y.copy()
    yt[4] = 3.0
    yi = djs_maskinterp1(y, np.array([0, 1, 0, 0, 1]))
    assert np.allclose(yt, yi)
    yi = djs_maskinterp1(y, np.array([0, 1, 0, 0, 1]), const=True)
    assert np.allclose(yt, yi)
    yt = y.copy()
    yt[0] = 1.0
    yi = djs_maskinterp1(y, np.array([1, 0, 1, 0, 0]), const=True)
    assert np.allclose(yt, yi)
    #
    # Test other x values
    #
    x = np.array([0.0, 0.5, 1.0, 1.5, 2.0], dtype=y.dtype)
    yi = djs_maskinterp1(y, np.array([0, 1, 0, 1, 0]), xval=x)
    assert np.allclose(y, yi)
    yt = y.copy()
    yt[4] = 3.0
    yi = djs_maskinterp1(y, np.array([0, 1, 0, 0, 1]), xval=x)
    assert np.allclose(yt, yi)
    yi = djs_maskinterp1(y, np.array([0, 1, 0, 0, 1]), xval=x, const=True)
    assert np.allclose(yt, yi)
    yt = y.copy()
    yt[0] = 1.0
    yi = djs_maskinterp1(y, np.array([1, 0, 1, 0, 0]), xval=x, const=True)
    assert np.allclose(yt, yi)


def test_djs_maskinterp():
    y = np.array([0.0, 1.0, 2.0, 3.0, 4.0], dtype=np.float64)
    mask = np.array([0, 1, 0])
    with pytest.raises(ValueError):
        yi = djs_maskinterp(y, mask)
    mask = np.array([0, 1, 0, 0, 0])
    x = np.array([0.0, 0.5, 1.0], dtype=y.dtype)
    with pytest.raises(ValueError):
        yi = djs_maskinterp(y, mask, xval=x)
    # 1-D case
    yi = djs_maskinterp(y, mask)
    assert np.allclose(y, yi)
    # 2-D case
    x = np.array([0.0, 0.5, 1.0, 1.5, 2.0], dtype=y.dtype)
    x = np.vstack((x, x, x))
    y = np.vstack((y, y, y))
    mask = np.vstack((mask, mask, mask))
    with pytest.raises(ValueError):
        yi = djs_maskinterp(y, mask)
    with pytest.raises(ValueError):
        yi = djs_maskinterp(y, mask, axis=-1)
    with pytest.raises(ValueError):
        yi = djs_maskinterp(y, mask, axis=2)
    yi = djs_maskinterp(y, mask, axis=0)
    assert np.allclose(y, yi)
    yi = djs_maskinterp(y, mask, axis=0, xval=x)
    assert np.allclose(y, yi)
    mask[:, 1] = 0
    mask[1, :] = 1
    yi = djs_maskinterp(y, mask, axis=1)
    assert np.allclose(y, yi)
    yi = djs_maskinterp(y, mask, axis=1, xval=x)
    assert np.allclose(y, yi)
    # 3-D case
    x = np.dstack((x, x, x, x, x, x, x))
    y = np.dstack((y, y, y, y, y, y, y))
    mask = np.dstack((mask, mask, mask, mask, mask, mask, mask))
    mask[:, :, :] = 0
    mask[:, :, 5] = 1
    yi = djs_maskinterp(y, mask, axis=0)
    assert np.allclose(y, yi)
    yi = djs_maskinterp(y, mask, axis=0, xval=x)
    assert np.allclose(y, yi)
    mask[:, :, :] = 0
    mask[:, 3, :] = 1
    yi = djs_maskinterp(y, mask, axis=1)
    assert np.allclose(y, yi)
    yi = djs_maskinterp(y, mask, axis=1, xval=x)
    assert np.allclose(y, yi)
    mask[:, :, :] = 0
    mask[1, :, :] = 1
    yi = djs_maskinterp(y, mask, axis=2)
    assert np.allclose(y, yi)
    yi = djs_maskinterp(y, mask, axis=2, xval=x)
    assert np.allclose(y, yi)
    # 4-D case
    y = np.random.random((2, 2, 2, 2))
    with pytest.raises(ValueError):
        yi = djs_maskinterp(y, (y > 0.5), axis=0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the rgbcolor directory in idlutils.

This code reproduces the algorithms of Lupton *et al.* (2004) [2]_.
The purpose is to produce nice color (RGB, JPEG, PNG, etc.) from FITS
images in astronomical filter bands.

References
----------

.. [2] `Lupton, Robert, et al., 2004 PASP 116, 113
   <https://ui.adsabs.harvard.edu/abs/2004PASP..116..133L/abstract>`_.
"""
import numpy as np
from warnings import warn
from . import PydlutilsUserWarning


def nw_arcsinh(colors, nonlinearity=3.0):
    """Scales `colors` by a degree of nonlinearity specified by user.

    The input image must have zero background (*i.e.*, it must already
    be background-subtracted).

    Parameters
    ----------
    colors : :class:`~numpy.ndarray`
        3D Array containing RGB image.  The dimensions should be (X, Y, 3).
    nonlinearity : :class:`float`
        Amount of nonlinear scaling.  If set to zero, no scaling will be
        performed (this is equivalent to linear scaling).

    Returns
    -------
    :class:`~numpy.ndarray`
        The scaled image.

    Raises
    ------
    :exc:`ValueError`
        If `colors` has the wrong shape.
    """
    if nonlinearity == 0:
        return colors
    if colors.ndim != 3:
        raise ValueError("A 3D image is required!")
    nx, ny, nc = colors.shape
    if nc != 3:
        raise ValueError("The 3D image must have 3 image planes.")
    radius = colors.sum(2)
    w = radius == 0
    radius[w] = radius[w] + 1.0
    fac = np.arcsinh(radius*nonlinearity)/nonlinearity/radius
    fitted_colors = np.zeros(colors.shape, dtype=colors.dtype)
    for k in range(3):
        fitted_colors[:, :, k] = colors[:, :, k]*fac
    return fitted_colors


def nw_cut_to_box(colors, origin=(0.0, 0.0, 0.0)):
    """Limits the pixel values of the image to a 'box', so that the colors
    do not saturate to white but to a specific color.

    Parameters
    ----------
    colors : :class:`~numpy.ndarray`
        3D Array containing RGB image.  The dimensions should be (X, Y, 3).
    origin : :class:`tuple` or :class:`~numpy.ndarray`
        An array with 3 elements.  The "distance" from this origin is
        considered saturated.

    Returns
    -------
    :class:`~numpy.ndarray`
        The "boxed" image.

    Raises
    ------
    :exc:`ValueError`
        If `colors` or `origin` has the wrong shape.
    """
    if len(origin) != 3:
        raise ValueError("The origin array must contain 3 elements!")
    if colors.ndim != 3:
        raise ValueError("A 3D image is required!")
    nx, ny, nc = colors.shape
    if nc != 3:
        raise ValueError("The 3D image must have 3 image planes.")
    pos_dist = 1.0 - np.array(origin, dtype=colors.dtype)
    factors = np.zeros(colors.shape, dtype=colors.dtype)
    for k in range(3):
        factors[:, :, k] = colors[:, :, k]/pos_dist[k]
    factor = factors.max(2)
    factor = np.where(factor > 1.0, factor, 1.0)
    boxed_colors = np.zeros(colors.shape, dtype=colors.dtype)
    for k in range(3):
        boxed_colors[:, :, k] = origin[k] + colors[:, :, k]/factor
    return boxed_colors


def nw_float_to_byte(image, bits=8):
    """Converts an array of floats in [0.0, 1.0] to bytes in [0, 255].

    Parameters
    ----------
    image : :class:`~numpy.ndarray`
        Image to convert.
    bits : :class:`int`, optional
        Number of bits in final image.

    Returns
    -------
    :class:`~numpy.ndarray`
        Converted image.
    """
    if bits > 8:
        warn("bits > 8 not supported; setting bits = 8.", PydlutilsUserWarning)
        bits = 8
    fmax = 1 << bits
    bmax = fmax - 1
    f1 = np.floor(image * fmax)
    f2 = np.where(f1 > 0, f1, 0)
    f3 = np.where(f2 < bmax, f2, bmax)
    return f3.astype(np.uint8)


def nw_scale_rgb(colors, scales=(1.0, 1.0, 1.0)):
    """Multiply RGB image by color-dependent scale factor.

    Parameters
    ----------
    colors : :class:`~numpy.ndarray`
        3D Array containing RGB image.  The dimensions should be (X, Y, 3).
    scales : :class:`tuple` or :class:`~numpy.ndarray`, optional
        An array with 3 elements.

    Returns
    -------
    :class:`~numpy.ndarray`
        The scaled image.

    Raises
    ------
    :exc:`ValueError`
        If `colors` or `scales` has the wrong shape.
    """
    if len(scales) != 3:
        raise ValueError("The scale factor must contain 3 elements!")
    if colors.ndim != 3:
        raise ValueError("A 3D image is required!")
    nx, ny, nc = colors.shape
    if nc != 3:
        raise ValueError("The 3D image must have 3 image planes.")
    scaled_colors = np.zeros(colors.shape, dtype=colors.dtype)
    for k in range(3):
        scaled_colors[:, :, k] = colors[:, :, k]*scales[k]
    return scaled_colors



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.math.
"""
import pytest
import numpy as np
from astropy.utils.data import get_pkg_data_filename
from ..math import computechi2, djs_median, find_contiguous


def test_computechi2():
    x = np.arange(20)
    y = np.array([6.6198438, 1.3491303, 0.41035045, 9.4013375, 4.1103360,
                  4.3522868, 4.6338078, 4.7400367, 5.1860726, 5.1082748,
                  5.1084596, 5.2990997, 5.5987537, 5.7007504, 5.7855296,
                  4.1123709, 6.9437957, 4.9956179, 4.3724215, 3.6245063])
    sqivar = np.array([1.32531, 0.886090, 1.08384, 1.04489, 1.46807,
                       1.30800, 0.507725, 1.12840, 0.955025, 1.35925,
                       1.10126, 1.45690, 0.575700, 0.949710, 1.23368,
                       0.536489, 0.772543, 0.957729, 0.883976, 1.11559])
    templates = np.vstack((np.ones((20,), dtype='d'), x)).transpose()
    chi2 = computechi2(y, sqivar, templates)
    #
    # 20 data points, 2 parameters = 18 degrees of freedom.
    #
    assert chi2.dof == 18
    #
    # The covariance matrix should be symmetric
    #
    assert (chi2.covar.T == chi2.covar).all()
    #
    # The variance vector should be the same as the diagonal of the
    # covariance matrix.
    #
    assert (chi2.var == np.diag(chi2.covar)).all()


def test_djs_median():
    test_data_file = get_pkg_data_filename('t/median_data.txt')
    test_data = np.loadtxt(test_data_file, dtype='d', delimiter=',')
    np.random.seed(424242)
    data = 100.0*np.random.random(100)
    data2 = 100.0*np.random.random((10, 10))
    data3 = 100.0*np.random.random((10, 10, 10))
    data_width_5 = test_data[0, :]
    # data_width_5_reflect = test_data[1, :]
    #
    # Degenerate cases that fall back on numpy.median().
    #
    assert np.allclose(np.median(data), djs_median(data))
    assert np.allclose(np.median(data2, axis=0),
                       djs_median(data2, dimension=0))
    assert np.allclose(np.median(data2, axis=1),
                       djs_median(data2, dimension=1))
    #
    # Test widths.
    #
    assert np.allclose(data, djs_median(data, width=1))
    assert np.allclose(data_width_5, djs_median(data, width=5))
    # assert np.allclose(data_width_5_reflect,
    #                    djs_median(data, width=5, boundary='reflect'))
    #
    # Exceptions
    #
    with pytest.raises(ValueError):
        _ = djs_median(data, width=5, boundary='nearest')
    with pytest.raises(ValueError):
        _ = djs_median(data, width=5, boundary='wrap')
    with pytest.raises(ValueError):
        _ = djs_median(data, width=5, boundary='foobar')
    with pytest.raises(ValueError):
        _ = djs_median(data2, width=5, dimension=1)
    with pytest.raises(ValueError):
        _ = djs_median(data3, width=5)
    with pytest.raises(ValueError):
        _ = djs_median(data3, width=5, boundary='reflect')


def test_find_contiguous():
    assert (find_contiguous(np.array([0, 1, 1, 1, 0, 1, 1, 0, 1])) ==
            [1, 2, 3])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the image directory in idlutils.
"""
import numpy as np


def djs_maskinterp1(yval, mask, xval=None, const=False):
    """Interpolate over a masked, 1-d array.

    Parameters
    ----------
    yval : :class:`numpy.ndarray`
        The input values.
    mask : :class:`numpy.ndarray`
        The mask.
    xval : :class:`numpy.ndarray`, optional
        If set, use these x values, otherwise use an array.
    const : :class:`bool`, optional
        If set to ``True``, bad values around the edges of the array will be
        set to a constant value.  Because of the default behavior of
        :func:`numpy.interp`, this value actually makes no difference in
        the output.

    Returns
    -------
    :class:`numpy.ndarray`
        The `yval` array with masked values replaced by interpolated values.
    """
    good = mask == 0
    if good.all():
        return yval
    ngood = good.sum()
    igood = good.nonzero()[0]
    if ngood == 0:
        return yval
    if ngood == 1:
        return np.zeros(yval.shape, dtype=yval.dtype) + yval[igood[0]]
    ynew = yval.astype('d')
    ny = yval.size
    ibad = (mask != 0).nonzero()[0]
    if xval is None:
        ynew[ibad] = np.interp(ibad, igood, ynew[igood])
        if const:
            if igood[0] != 0:
                ynew[0:igood[0]] = ynew[igood[0]]
            if igood[ngood-1] != ny-1:
                ynew[igood[ngood-1]+1:ny] = ynew[igood[ngood-1]]
    else:
        ii = xval.argsort()
        ibad = (mask[ii] != 0).nonzero()[0]
        igood = (mask[ii] == 0).nonzero()[0]
        ynew[ii[ibad]] = np.interp(xval[ii[ibad]], xval[ii[igood]],
                                   ynew[ii[igood]])
        if const:
            if igood[0] != 0:
                ynew[ii[0:igood[0]]] = ynew[ii[igood[0]]]
            if igood[ngood-1] != ny-1:
                ynew[ii[igood[ngood-1]+1:ny]] = ynew[ii[igood[ngood-1]]]
    return ynew


def djs_maskinterp(yval, mask, xval=None, axis=None, const=False):
    """Interpolate over masked pixels in a vector, image or 3-D array.

    Parameters
    ----------
    yval : :class:`numpy.ndarray`
        The input values
    mask : :class:`numpy.ndarray`
        The mask
    xval : :class:`numpy.ndarray`, optional
        If set, use these x values, otherwise use an array
    axis : :class:`int`, optional
        Must be set if yval has more than one dimension. If set to zero,
        interpolate along the first axis of the array, if set to one,
        interpolate along the second axis of the array, and so on.
    const : :class:`bool`, optional
        This value is passed to a helper function, djs_maskinterp1.

    Returns
    -------
    :class:`numpy.ndarray`
        The interpolated array.
    """
    if mask.shape != yval.shape:
        raise ValueError('mask must have the same shape as yval.')
    if xval is not None:
        if xval.shape != yval.shape:
            raise ValueError('xval must have the same shape as yval.')
    ndim = yval.ndim
    if ndim == 1:
        ynew = djs_maskinterp1(yval, mask, xval=xval, const=const)
    else:
        if axis is None:
            raise ValueError('Must set axis if yval has more than one dimension.')
        if axis < 0 or axis > ndim-1 or axis - int(axis) != 0:
            raise ValueError('Invalid axis value.')
        ynew = np.zeros(yval.shape, dtype=yval.dtype)
        if ndim == 2:
            if xval is None:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        ynew[i, :] = djs_maskinterp1(yval[i, :], mask[i, :],
                                                     const=const)
                else:
                    for i in range(yval.shape[1]):
                        ynew[:, i] = djs_maskinterp1(yval[:, i], mask[:, i],
                                                     const=const)
            else:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        ynew[i, :] = djs_maskinterp1(yval[i, :], mask[i, :],
                                                     xval=xval[i, :],
                                                     const=const)
                else:
                    for i in range(yval.shape[1]):
                        ynew[:, i] = djs_maskinterp1(yval[:, i], mask[:, i],
                                                     xval=xval[:, i],
                                                     const=const)
        elif ndim == 3:
            if xval is None:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[1]):
                            ynew[i, j, :] = djs_maskinterp1(yval[i, j, :],
                                                            mask[i, j, :],
                                                            const=const)
                elif axis == 1:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[2]):
                            ynew[i, :, j] = djs_maskinterp1(yval[i, :, j],
                                                            mask[i, :, j],
                                                            const=const)
                else:
                    for i in range(yval.shape[1]):
                        for j in range(yval.shape[2]):
                            ynew[:, i, j] = djs_maskinterp1(yval[:, i, j],
                                                            mask[:, i, j],
                                                            const=const)
            else:
                if axis == 0:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[1]):
                            ynew[i, j, :] = djs_maskinterp1(yval[i, j, :],
                                                            mask[i, j, :],
                                                            xval=xval[i, j, :],
                                                            const=const)
                elif axis == 1:
                    for i in range(yval.shape[0]):
                        for j in range(yval.shape[2]):
                            ynew[i, :, j] = djs_maskinterp1(yval[i, :, j],
                                                            mask[i, :, j],
                                                            xval=xval[i, :, j],
                                                            const=const)
                else:
                    for i in range(yval.shape[1]):
                        for j in range(yval.shape[2]):
                            ynew[:, i, j] = djs_maskinterp1(yval[:, i, j],
                                                            mask[:, i, j],
                                                            xval=xval[:, i, j],
                                                            const=const)
        else:
            raise ValueError('Unsupported number of dimensions.')
    return ynew



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the spheregroup directory in idlutils.
"""
from warnings import warn
import numpy as np
from . import PydlutilsException, PydlutilsUserWarning
from ..goddard.astro import gcirc


class chunks(object):
    """chunks class

    Functions for creating and manipulating spherical chunks are implemented
    as methods on this class.
    """

    def __init__(self, ra, dec, minSize):
        """Init creates an object whose attributes are similar those created
        by the setchunks() function in the spheregroup library.
        """
        #
        # Save the value of minSize
        #
        self.minSize = minSize
        #
        # Find maximum and minimum dec (in degrees)
        #
        decMin = dec.min()
        decMax = dec.max()
        decRange = decMax - decMin
        #
        # Find the declination boundaries; make them an integer multiple of
        # minSize, with extra room (one cell) on the edges.
        #
        self.nDec = 3 + int(np.floor(decRange/minSize))
        decRange = minSize*float(self.nDec)
        decMin = decMin - 0.5*(decRange - decMax + decMin)
        decMax = decMin + decRange
        if decMin < -90.0 + 3.0*minSize:
            decMin = -90.0
        if decMax > 90.0 - 3.0*minSize:
            decMax = 90.0
        self.decBounds = decMin + ((decMax - decMin) * np.arange(self.nDec + 1, dtype='d'))/float(self.nDec)
        #
        # Find ra offset which minimizes the range in ra (this should take care
        # of the case that ra crosses zero in some parts
        #
        if abs(self.decBounds[self.nDec]) > abs(self.decBounds[0]):
            cosDecMin = np.cos(np.deg2rad(self.decBounds[self.nDec]))
        else:
            cosDecMin = np.cos(np.deg2rad(self.decBounds[0]))
        if cosDecMin <= 0.0:
            raise PydlutilsException("cosDecMin={0:f} not positive in setchunks().".format(cosDecMin))
        self.raRange, self.raOffset = self.rarange(ra, minSize/cosDecMin)
        self.raMin, self.raMax = self.getraminmax(ra, self.raOffset)
        #
        # Isn't this redundant?
        #
        self.raRange = self.raMax - self.raMin
        #
        # For each declination slice, find the number of ra divisions
        # necessary and set them
        #
        self.raBounds = list()
        self.nRa = list()
        for i in range(self.nDec):
            #
            # Get maximum declination and its cosine
            #
            if abs(self.decBounds[i]) > abs(self.decBounds[i+1]):
                cosDecMin = np.cos(np.deg2rad(self.decBounds[i]))
            else:
                cosDecMin = np.cos(np.deg2rad(self.decBounds[i+1]))
            if cosDecMin <= 0.0:
                raise PydlutilsException("cosDecMin={0:f} not positive in setchunks().".format(cosDecMin))
            #
            # Get raBounds array for this declination array, leave an extra
            # cell on each end
            #
            self.nRa.append(3 + int(np.floor(cosDecMin*self.raRange/minSize)))
            raRangeTmp = minSize*float(self.nRa[i])/cosDecMin
            raMinTmp = self.raMin - 0.5*(raRangeTmp-self.raMax+self.raMin)
            raMaxTmp = raMinTmp + raRangeTmp
            #
            # If we cannot avoid the 0/360 point, embrace it
            #
            if (raRangeTmp >= 360.0 or
                    raMinTmp <= minSize/cosDecMin or
                    raMaxTmp >= 360.0 - minSize/cosDecMin or
                    abs(self.decBounds[i]) == 90.0):
                raMinTmp = 0.0
                raMaxTmp = 360.0
                raRangeTmp = 360.0
            if self.decBounds[i] == -90.0 or self.decBounds[i+1] == 90.0:
                self.nRa[i] = 1
            self.raBounds.append(raMinTmp +
                                 (raMaxTmp - raMinTmp) * np.arange(self.nRa[i] + 1, dtype='d') /
                                 float(self.nRa[i]))
        #
        # Create an empty set of lists to hold the output of self.assign()
        #
        self.chunkList = [[list() for j in range(self.nRa[i])] for i in range(self.nDec)]
        #
        # nChunkMax will be the length of the largest list in chunkList
        # it is computed by chunks.assign()
        #
        self.nChunkMax = 0
        return

    def rarange(self, ra, minSize):
        """Finds the offset which yields the smallest raRange & returns both.

        Notes
        -----

        .. warning:: This is not (yet) well-defined for the case of only one point.
        """
        NRA = 6
        raRangeMin = 361.0
        raOffset = 0.0
        EPS = 1.0e-5
        for j in range(NRA):
            raMin, raMax = self.getraminmax(ra, 360.0*float(j)/float(NRA))
            raRange = raMax-raMin
            if (2.0*(raRange-raRangeMin)/(raRange+raRangeMin) < -EPS and
                    raMin > minSize and raMax < 360.0 - minSize):
                raRangeMin = raRange
                raOffset = 360.0*float(j)/float(NRA)
        return (raRangeMin, raOffset)

    def getraminmax(self, ra, raOffset):
        """Utility function used by rarange.
        """
        currRa = np.fmod(ra + raOffset, 360.0)
        return (currRa.min(), currRa.max())

    def cosDecMin(self, i):
        """Frequently used utility function.
        """
        if abs(self.decBounds[i]) > abs(self.decBounds[i+1]):
            return np.cos(np.deg2rad(self.decBounds[i]))
        else:
            return np.cos(np.deg2rad(self.decBounds[i+1]))

    def assign(self, ra, dec, marginSize):
        """Take the objects and the chunks (already defined in the constructor)
        and assign the objects to the appropriate chunks, with some leeway
        given by the parameter marginSize.  Basically, at the end, each
        chunk should be associated with a list of the objects that belong
        to it.
        """
        if marginSize >= self.minSize:
            raise PydlutilsException("marginSize>=minSize ({0:f}={1:f}) in chunks.assign().".format(marginSize, self.minSize))
        chunkDone = [[False for j in range(self.nRa[i])] for i in range(self.nDec)]
        for i in range(ra.size):
            currRa = np.fmod(ra[i] + self.raOffset, 360.0)
            try:
                raChunkMin, raChunkMax, decChunkMin, decChunkMax = self.getbounds(currRa, dec[i], marginSize)
            except PydlutilsException:
                continue
            #
            # Reset chunkDone.  This is silly, but is necessary to
            # reproduce the logic.
            #
            for decChunk in range(decChunkMin, decChunkMax+1):
                for raChunk in range(raChunkMin[decChunk-decChunkMin]-1, raChunkMax[decChunk-decChunkMin]+2):
                    if raChunk < 0:
                        currRaChunk = (raChunk+self.nRa[decChunk]) % self.nRa[decChunk]
                    elif raChunk > self.nRa[decChunk]-1:
                        currRaChunk = (raChunk-self.nRa[decChunk]) % self.nRa[decChunk]
                    else:
                        currRaChunk = raChunk
                    if currRaChunk >= 0 and currRaChunk <= self.nRa[decChunk]-1:
                        chunkDone[decChunk][currRaChunk] = False
            for decChunk in range(decChunkMin, decChunkMax+1):
                for raChunk in range(raChunkMin[decChunk-decChunkMin], raChunkMax[decChunk-decChunkMin]+1):
                    if raChunk < 0:
                        currRaChunk = (raChunk+self.nRa[decChunk]) % self.nRa[decChunk]
                    elif raChunk > self.nRa[decChunk]-1:
                        currRaChunk = (raChunk-self.nRa[decChunk]) % self.nRa[decChunk]
                    else:
                        currRaChunk = raChunk
                    if currRaChunk >= 0 and currRaChunk <= self.nRa[decChunk]-1:
                        if not chunkDone[decChunk][currRaChunk]:
                            self.chunkList[decChunk][currRaChunk].append(i)
                            #
                            # Update nChunkMax
                            #
                            if len(self.chunkList[decChunk][currRaChunk]) > self.nChunkMax:
                                self.nChunkMax = len(self.chunkList[decChunk][currRaChunk])
                            chunkDone[decChunk][currRaChunk] = True
        return

    def getbounds(self, ra, dec, marginSize):
        """Find the set of chunks a point (with margin) belongs to.
        """
        #
        # Find the declination slice without regard to marginSize
        #
        decChunkMin = int(np.floor((dec - self.decBounds[0]) *
                                   float(self.nDec) /
                                   (self.decBounds[self.nDec]-self.decBounds[0])))
        decChunkMax = decChunkMin
        if decChunkMin < 0 or decChunkMin > self.nDec - 1:
            raise PydlutilsException("decChunkMin out of range in chunks.getbounds().")
        #
        # Set minimum and maximum bounds of dec
        #
        while dec - self.decBounds[decChunkMin] < marginSize and decChunkMin > 0:
            decChunkMin -= 1
        while self.decBounds[decChunkMax+1] - dec < marginSize and decChunkMax < self.nDec - 1:
            decChunkMax += 1
        #
        # Find ra chunk bounds for each dec chunk
        #
        raChunkMin = np.zeros(decChunkMax-decChunkMin+1, dtype='i4')
        raChunkMax = np.zeros(decChunkMax-decChunkMin+1, dtype='i4')
        for i in range(decChunkMin, decChunkMax+1):
            cosDecMin = self.cosDecMin(i)
            raChunkMin[i-decChunkMin] = int(np.floor((ra - self.raBounds[i][0]) *
                                                     float(self.nRa[i]) /
                                                     (self.raBounds[i][self.nRa[i]] - self.raBounds[i][0])))
            raChunkMax[i-decChunkMin] = raChunkMin[i-decChunkMin]
            if raChunkMin[i-decChunkMin] < 0 or raChunkMin[i-decChunkMin] > self.nRa[i]-1:
                raise PydlutilsException("raChunkMin out of range in chunks.getbounds().")
            #
            # Set minimum and maximum bounds of ra
            #
            raCheck = raChunkMin[i-decChunkMin]
            keepGoing = True
            while keepGoing and raCheck > -1:
                if raCheck >= 0 and raCheck < self.nRa[i]:
                    keepGoing = (ra - self.raBounds[i][raCheck])*cosDecMin < marginSize
                else:
                    keepGoing = False
                if keepGoing:
                    raCheck -= 1
            raChunkMin[i-decChunkMin] = raCheck
            raCheck = raChunkMax[i-decChunkMin]
            keepGoing = True
            while keepGoing and raCheck < self.nRa[i]:
                if raCheck >= 0 and raCheck < self.nRa[i]:
                    keepGoing = (self.raBounds[i][raCheck+1]-ra)*cosDecMin < marginSize
                else:
                    keepGoing = False
                if keepGoing:
                    raCheck += 1
            raChunkMax[i-decChunkMin] = raCheck
        return (raChunkMin, raChunkMax, decChunkMin, decChunkMax)

    def get(self, ra, dec):
        """Find the chunk to which a given point belongs.
        """
        #
        # Find dec chunk
        #
        decChunk = int(np.floor((dec - self.decBounds[0]) *
                                float(self.nDec) /
                                (self.decBounds[self.nDec]-self.decBounds[0])))
        #
        # Find ra chunk
        #
        if decChunk < self.nDec and decChunk >= 0:
            raChunk = int(np.floor((ra - self.raBounds[decChunk][0]) *
                                   float(self.nRa[decChunk]) /
                                   (self.raBounds[decChunk][self.nRa[decChunk]] - self.raBounds[decChunk][0])))
            if raChunk < 0 or raChunk > self.nRa[decChunk]-1:
                raise PydlutilsException("raChunk out of range in chunks.get()")
        else:
            raChunk = -1
        return (raChunk, decChunk)

    def friendsoffriends(self, ra, dec, linkSep):
        """Friends-of-friends using chunked data.
        """
        nPoints = ra.size
        inGroup = np.zeros(nPoints, dtype='i4') - 1
        #
        # mapGroups contains an equivalency mapping of groups.  mapGroup[i]=j
        # means i and j are actually the same group.  j<=i always, by design.
        # The largest number of groups you can get
        # (assuming linkSep < marginSize < minSize) is 9 times the number of
        # targets
        #
        mapGroups = np.zeros(9*nPoints, dtype='i4') - 1
        nMapGroups = 0
        for i in range(self.nDec):
            for j in range(self.nRa[i]):
                if len(self.chunkList[i][j]) > 0:
                    chunkGroup = self.chunkfriendsoffriends(ra, dec, self.chunkList[i][j], linkSep)
                    for k in range(chunkGroup.nGroups):
                        minEarly = 9*nPoints
                        l = chunkGroup.firstGroup[k]
                        while l != -1:
                            if inGroup[self.chunkList[i][j][l]] != -1:
                                checkEarly = inGroup[self.chunkList[i][j][l]]
                                while mapGroups[checkEarly] != checkEarly:
                                    checkEarly = mapGroups[checkEarly]
                                minEarly = min(minEarly, checkEarly)
                            else:
                                inGroup[self.chunkList[i][j][l]] = nMapGroups
                            l = chunkGroup.nextGroup[l]
                        if minEarly == 9*nPoints:
                            mapGroups[nMapGroups] = nMapGroups
                        else:
                            mapGroups[nMapGroups] = minEarly
                            l = chunkGroup.firstGroup[k]
                            while l != -1:
                                checkEarly = inGroup[self.chunkList[i][j][l]]
                                while mapGroups[checkEarly] != checkEarly:
                                    tmpEarly = mapGroups[checkEarly]
                                    mapGroups[checkEarly] = minEarly
                                    checkEarly = tmpEarly
                                mapGroups[checkEarly] = minEarly
                                l = chunkGroup.nextGroup[l]
                        nMapGroups += 1
        #
        # Now all groups which are mapped to themselves are the real groups
        # Make sure the mappings are set up to go all the way down.
        #
        nGroups = 0
        for i in range(nMapGroups):
            if mapGroups[i] != -1:
                if mapGroups[i] == i:
                    mapGroups[i] = nGroups
                    nGroups += 1
                else:
                    mapGroups[i] = mapGroups[mapGroups[i]]
            else:
                raise PydlutilsException("MapGroups[{0:d}]={1:d} in chunks.friendsoffriends().".format(i, mapGroups[i]))
        for i in range(nPoints):
            inGroup[i] = mapGroups[inGroup[i]]
        firstGroup = np.zeros(nPoints, dtype='i4') - 1
        nextGroup = np.zeros(nPoints, dtype='i4') - 1
        multGroup = np.zeros(nPoints, dtype='i4')
        for i in range(nPoints-1, -1, -1):
            nextGroup[i] = firstGroup[inGroup[i]]
            firstGroup[inGroup[i]] = i
        for i in range(nGroups):
            j = firstGroup[i]
            while j != -1:
                multGroup[i] += 1
                j = nextGroup[j]
        return (inGroup, multGroup, firstGroup, nextGroup, nGroups)

    def chunkfriendsoffriends(self, ra, dec, chunkList, linkSep):
        """Does friends-of-friends on the ra, dec that are defined by
        chunkList.
        """
        #
        # Convert ra, dec into something that can be digested by the
        # groups object.
        #
        x = np.deg2rad(np.vstack((ra[chunkList], dec[chunkList])))
        radLinkSep = np.deg2rad(linkSep)
        group = groups(x, radLinkSep, 'sphereradec')
        return group


class groups(object):
    """Group a set of objects (a list of coordinates in some space) based on
    a friends-of-friends algorithm
    """

    @staticmethod
    def euclid(x1, x2):
        """Pythagorean theorem in Euclidean space with arbitrary number
        of dimensions.
        """
        return np.sqrt(((x1-x2)**2).sum())

    @staticmethod
    def sphereradec(x1, x2):
        """Separation of two points on a 2D-sphere, assuming they are in
        longitude-latitude or right ascension-declination form.  Assumes
        everything is already in radians.
        """
        return gcirc(x1[0], x1[1], x2[0], x2[1], units=0)

    def __init__(self, coordinates, distance, separation='euclid'):
        """Init creates an object and performs the friends-of-friends
        algorithm.  The coordinates can have arbitrary dimensions, with each
        column representing one of the dimensions.  Each row defines an object.
        If separation is not defined it defaults to Euclidean space.
        """
        #
        # Find a separation function
        #
        if callable(separation):
            self.separation = separation
        elif isinstance(separation, (str,)):
            if separation == 'euclid':
                self.separation = self.euclid
            elif separation == 'sphereradec':
                self.separation = self.sphereradec
            else:
                raise PydlutilsException("Unknown separation function: {0}.".format(separation))
        else:
            raise PydlutilsException("Improper type for separation!")
        #
        # Save information about the coordinates.
        #
        nGroups = 0
        nTargets = coordinates.shape[1]
        multGroup = np.zeros(nTargets, dtype='i4')
        firstGroup = np.zeros(nTargets, dtype='i4') - 1
        nextGroup = np.zeros(nTargets, dtype='i4') - 1
        inGroup = np.arange(nTargets, dtype='i4')
        #
        # Find all the other targets associated with each target
        #
        for i in range(nTargets):
            nTmp = 0
            minGroup = nGroups
            for j in range(nTargets):
                sep = self.separation(coordinates[:, i], coordinates[:, j])
                if sep <= distance:
                    multGroup[nTmp] = j
                    minGroup = min(minGroup, inGroup[j])
                    nTmp += 1
            #
            # Use this minimum for all
            #
            for j in range(nTmp):
                if inGroup[multGroup[j]] < nTargets:
                    k = firstGroup[inGroup[multGroup[j]]]
                    while k != -1:
                        inGroup[k] = minGroup
                        k = nextGroup[k]
                inGroup[multGroup[j]] = minGroup
            #
            # If it is a new group (no earlier groups), increment nGroups
            #
            if minGroup == nGroups:
                nGroups += 1
            for j in range(i+1):
                firstGroup[j] = -1
            for j in range(i, -1, -1):
                nextGroup[j] = firstGroup[inGroup[j]]
                firstGroup[inGroup[j]] = j
        #
        # Renumber to get rid of the numbers which were skipped
        #
        renumbered = np.zeros(nTargets, dtype='bool')
        nTmp = nGroups
        nGroups = 0
        for i in range(nTargets):
            if not renumbered[i]:
                j = firstGroup[inGroup[i]]
                while j != -1:
                    inGroup[j] = nGroups
                    renumbered[j] = True
                    j = nextGroup[j]
                nGroups += 1
        #
        # Reset the values of firstGroup and inGroup
        #
        firstGroup[:] = -1
        for i in range(nTargets-1, -1, -1):
            nextGroup[i] = firstGroup[inGroup[i]]
            firstGroup[inGroup[i]] = i
        #
        # Get the multiplicity
        #
        for i in range(nGroups):
            multGroup[i] = 0
            j = firstGroup[i]
            while j != -1:
                multGroup[i] += 1
                j = nextGroup[j]
        #
        # Set attributes
        #
        self.nGroups = nGroups
        self.nTargets = nTargets
        self.inGroup = inGroup
        self.multGroup = multGroup
        self.firstGroup = firstGroup
        self.nextGroup = nextGroup
        return


def spheregroup(ra, dec, linklength, chunksize=None):
    """Perform friends-of-friends grouping given ra/dec coordinates.

    Parameters
    ----------
    ra, dec : :class:`numpy.ndarray`
        Arrays of coordinates to group in decimal degrees.
    linklength : :class:`float`
        Linking length for the groups in decimal degrees.
    chunksize : :class:`float`, optional
        Break up the sphere into chunks of this size in decimal degrees.

    Returns
    -------
    :class:`tuple`
        A tuple containing the group number of each object, the multiplicity
        of each group, the first member of each group, and the next
        member of the group for each object.

    Raises
    ------
    :exc:`PydlutilsException`
        If the array of coordinates only contains one point.

    Notes
    -----
    It is important that `chunksize` >= 4 * `linklength`.  This is enforced.

    .. warning:: Behavior at the poles is not well tested.
    """
    npoints = ra.size
    if npoints == 1:
        raise PydlutilsException("Cannot group only one point!")
    #
    # Define the chunksize
    #
    if chunksize is not None:
        if chunksize < 4.0*linklength:
            chunksize = 4.0*linklength
            warn("chunksize changed to {0:.2f}.".format(chunksize), PydlutilsUserWarning)
    else:
        chunksize = max(4.0*linklength, 0.1)
    #
    # Initialize chunks
    #
    chunk = chunks(ra, dec, chunksize)
    chunk.assign(ra, dec, linklength)
    #
    # Run friends-of-friends
    #
    ingroup, multgroup, firstgroup, nextgroup, ngroups = chunk.friendsoffriends(ra, dec, linklength)
    #
    # Renumber the groups in order of appearance
    #
    renumbered = np.zeros(npoints, dtype='bool')
    iclump = 0
    for i in range(npoints):
        if not renumbered[i]:
            j = firstgroup[ingroup[i]]
            while j != -1:
                ingroup[j] = iclump
                renumbered[j] = True
                j = nextgroup[j]
            iclump += 1
    #
    # Reset the index lists
    #
    firstgroup[:] = -1
    for i in range(npoints-1, -1, -1):
        nextgroup[i] = firstgroup[ingroup[i]]
        firstgroup[ingroup[i]] = i
    #
    # Reset the multiplicities
    #
    multgroup[:] = 0
    for i in range(ngroups):
        j = firstgroup[i]
        while j != -1:
            multgroup[i] += 1
            j = nextgroup[j]
    return (ingroup, multgroup, firstgroup, nextgroup)


def spherematch(ra1, dec1, ra2, dec2, matchlength, chunksize=None,
                maxmatch=1):
    """Match points on a sphere.

    Parameters
    ----------
    ra1, dec1, ra2, dec2 : :class:`numpy.ndarray`
        The sets of coordinates to match.  Assumed to be in decimal degrees
    matchlength : :class:`float`
        Two points closer than this separation are matched. Assumed to be in decimal degrees.
    chunksize : :class:`float`, optional
        Value to pass to chunk assignment.
    maxmatch : :class:`int`, optional
        Allow up to `maxmatch` matches per coordinate.  Default 1. If set to zero,
        All possible matches will be returned.

    Returns
    -------
    :class:`tuple`
        A tuple containing the indices into the first set of points, the
        indices into the second set of points and the match distance in
        decimal degrees.

    Notes
    -----
    If you have sets of coordinates that differ in size, call this function
    with the larger list first.  This exploits the inherent asymmetry in the
    underlying code to reduce memory use.

    .. warning:: Behavior at the poles is not well tested.
    """
    #
    # Set default values
    #
    if chunksize is None:
        chunksize = max(4.0*matchlength, 0.1)
    #
    # Check input size
    #
    if ra1.size == 1:
        raise PydlutilsException("Change the order of the sets of coordinates!")
    #
    # Initialize chunks
    #
    chunk = chunks(ra1, dec1, chunksize)
    chunk.assign(ra2, dec2, matchlength)
    #
    # Create return arrays
    #
    match1 = list()
    match2 = list()
    distance12 = list()
    for i in range(ra1.size):
        currra = np.fmod(ra1[i]+chunk.raOffset, 360.0)
        rachunk, decchunk = chunk.get(currra, dec1[i])
        jmax = len(chunk.chunkList[decchunk][rachunk])
        if jmax > 0:
            for j in range(jmax):
                k = chunk.chunkList[decchunk][rachunk][j]
                sep = gcirc(ra1[i], dec1[i], ra2[k], dec2[k], units=2)/3600.0
                if sep < matchlength:
                    match1.append(i)
                    match2.append(k)
                    distance12.append(sep)
    #
    # Sort distances
    #
    omatch1 = np.array(match1)
    omatch2 = np.array(match2)
    odistance12 = np.array(distance12)
    s = odistance12.argsort()
    #
    # Retain only desired matches
    #
    if maxmatch > 0:
        gotten1 = np.zeros(ra1.size, dtype='i4')
        gotten2 = np.zeros(ra2.size, dtype='i4')
        nmatch = 0
        for i in range(omatch1.size):
            if (gotten1[omatch1[s[i]]] < maxmatch and
                    gotten2[omatch2[s[i]]] < maxmatch):
                gotten1[omatch1[s[i]]] += 1
                gotten2[omatch2[s[i]]] += 1
                nmatch += 1
        match1 = np.zeros(nmatch, dtype='i4')
        match2 = np.zeros(nmatch, dtype='i4')
        distance12 = np.zeros(nmatch, dtype='d')
        gotten1[:] = 0
        gotten2[:] = 0
        nmatch = 0
        for i in range(omatch1.size):
            if (gotten1[omatch1[s[i]]] < maxmatch and
                    gotten2[omatch2[s[i]]] < maxmatch):
                gotten1[omatch1[s[i]]] += 1
                gotten2[omatch2[s[i]]] += 1
                match1[nmatch] = omatch1[s[i]]
                match2[nmatch] = omatch2[s[i]]
                distance12[nmatch] = odistance12[s[i]]
                nmatch += 1
    else:
        match1 = omatch1[s]
        match2 = omatch2[s]
        distance12 = odistance12[s]
    return (match1, match2, distance12)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the goddard/astro directory in idlutils.
"""
from time import time
import numpy as np
from astropy.units import Angstrom


def airtovac(air):
    """Convert air wavelengths to wavelengths in vacuum.

    Parameters
    ----------
    air : array-like
        Values of wavelength in air in Angstroms.
        :class:`~astropy.units.Quantity` objects with valid length
        dimensions will be internally converted to Angstrom.

    Returns
    -------
    array-like
        Values of wavelength in vacuum in Angstroms.  If a
        :class:`~astropy.units.Quantity` object was passed in, the output
        will be converted to the same units as the input.

    Notes
    -----
    * Formula from `P. E. Ciddor, Applied Optics, 35, 1566 (1996)
      <https://ui.adsabs.harvard.edu/abs/1996ApOpt..35.1566C/abstract>`_.
    * Values of wavelength below 2000 Å are not converted.
    """
    try:
        u = air.unit
    except AttributeError:
        u = None
    try:
        t = air.dtype
    except AttributeError:
        # Most likely, air is simply a float.
        t = None
    if t is None:
        if air < 2000.0:
            return air
        vacuum = air
        a = air
        g = None
    else:
        try:
            a = air.to(Angstrom).value
        except AttributeError:
            a = air
        g = a < 2000.0
        if g.all():
            return air
        vacuum = np.zeros(air.shape, dtype=t) + a
    for k in range(2):
        sigma2 = (1.0e4/vacuum)**2
        fact = (1.0 + 5.792105e-2/(238.0185 - sigma2) +
                1.67917e-3/(57.362 - sigma2))
        vacuum = a * fact
    if g is not None:
        vacuum[g] = a[g]
    if u is not None:
        vacuum = (vacuum * Angstrom).to(u)
    return vacuum


def gcirc(ra1, dec1, ra2, dec2, units=2):
    """Computes rigorous great circle arc distances.

    Parameters
    ----------
    ra1, dec1, ra2, dec2 : :class:`float` or array-like
        RA and Dec of two points.
    units : { 0, 1, 2 }, optional
        * units = 0: everything is already in radians
        * units = 1: RA in hours, dec in degrees, distance in arcsec.
        * units = 2: RA, dec in degrees, distance in arcsec (default)

    Returns
    -------
    :class:`float` or array-like
        The angular distance.  Units of the value returned depend on the
        input value of `units`.

    Notes
    -----
    The formula below is the one best suited to handling small angular
    separations.  See:
    https://en.wikipedia.org/wiki/Great-circle_distance
    """
    if units == 0:
        rarad1 = ra1
        dcrad1 = dec1
        rarad2 = ra2
        dcrad2 = dec2
    elif units == 1:
        rarad1 = np.deg2rad(15.0*ra1)
        dcrad1 = np.deg2rad(dec1)
        rarad2 = np.deg2rad(15.0*ra2)
        dcrad2 = np.deg2rad(dec2)
    elif units == 2:
        rarad1 = np.deg2rad(ra1)
        dcrad1 = np.deg2rad(dec1)
        rarad2 = np.deg2rad(ra2)
        dcrad2 = np.deg2rad(dec2)
    else:
        raise ValueError('units must be 0, 1 or 2!')
    deldec2 = (dcrad2-dcrad1)/2.0
    delra2 = (rarad2-rarad1)/2.0
    sindis = np.sqrt(np.sin(deldec2)*np.sin(deldec2) +
                     np.cos(dcrad1)*np.cos(dcrad2)*np.sin(delra2)*np.sin(delra2))
    dis = 2.0*np.arcsin(sindis)
    if units == 0:
        return dis
    else:
        return np.rad2deg(dis)*3600.0


def get_juldate(seconds=None):
    """Returns the current Julian date.

    Uses the MJD trick & adds the offset to get JD.

    Parameters
    ----------
    seconds : :class:`int` or :class:`float`, optional
        Time in seconds since the UNIX epoch.  This should only be used
        for testing.

    Returns
    -------
    :class:`float`
        The Julian Day number as a floating point number.

    Notes
    -----
    Do not use this function if high precision is required, or if you are
    concerned about the distinction between UTC & TAI.
    """
    if seconds is None:
        t = time()
    else:
        t = seconds
    mjd = t/86400.0 + 40587.0
    return mjd + 2400000.5


def get_juldate_main():  # pragma: no cover
    """Entry point for the get_juldate command-line script.
    """
    print(get_juldate())
    return 0


def vactoair(vacuum):
    """Convert vacuum wavelengths to wavelengths in air.

    Parameters
    ----------
    vacuum : array-like
        Values of wavelength in vacuum in Angstroms.
        :class:`~astropy.units.Quantity` objects with valid length
        dimensions will be internally converted to Angstrom.

    Returns
    -------
    array-like
        Values of wavelength in air in Angstroms.
        :class:`~astropy.units.Quantity` object was passed in, the output
        will be converted to the same units as the input.

    Notes
    -----
    * Formula from `P. E. Ciddor, Applied Optics, 35, 1566 (1996)
      <https://ui.adsabs.harvard.edu/abs/1996ApOpt..35.1566C/abstract>`_.
    * Values of wavelength below 2000 Å are not converted.
    """
    try:
        u = vacuum.unit
    except AttributeError:
        u = None
    try:
        t = vacuum.dtype
    except AttributeError:
        # Most likely, vacuum is simply a float.
        t = None
    if t is None:
        if vacuum < 2000.0:
            return vacuum
        air = vacuum
        v = vacuum
        g = None
    else:
        try:
            v = vacuum.to(Angstrom).value
        except AttributeError:
            v = vacuum
        g = v < 2000.0
        if g.all():
            return vacuum
        air = np.zeros(vacuum.shape, dtype=t) + v
    sigma2 = (1.0e4/v)**2
    fact = (1.0 + 5.792105e-2/(238.0185 - sigma2) +
            1.67917e-3/(57.362 - sigma2))
    air = v / fact
    if g is not None:
        air[g] = v[g]
    if u is not None:
        air = (air * Angstrom).to(u)
    return air



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the pydl.goddard package.
"""
import pytest
import numpy as np
import astropy.units as u
from ..astro import airtovac, gcirc, get_juldate, vactoair
from ..math import flegendre
from ..misc import cirrange


def test_airtovac():
    vacuum = airtovac(1900.0)
    assert vacuum == 1900.0
    vacuum = airtovac(2000.0)
    assert np.allclose(vacuum, 2000.6475)
    air = np.array([1800.0, 1900.0, 2000.0, 2100.0, 2200.0, 2300.0])
    vacuum = airtovac(air)
    assert np.allclose(vacuum,
                       np.array([1800.0, 1900.0, 2000.6475,
                                 2100.6664, 2200.6868, 2300.7083]))
    vacuum = airtovac(6056.125)
    assert np.allclose(vacuum, 6057.8019)
    vacuum = airtovac(np.array([1800.0, 1850.0, 1900.0]))
    assert np.allclose(vacuum, np.array([1800.0, 1850.0, 1900.0]))
    #
    # Regression test for #8.
    #
    wave = air.reshape(2, 3)
    vacuum = airtovac(wave)
    assert np.allclose(vacuum,
                       np.array([[1800.0, 1900.0, 2000.6475],
                                 [2100.6664, 2200.6868, 2300.7083]]))
    #
    # Test with units.
    #
    air = air * u.Angstrom
    vacuum = airtovac(air)
    assert np.allclose(vacuum.value,
                       np.array([1800.0, 1900.0, 2000.6475,
                                 2100.6664, 2200.6868, 2300.7083]))
    assert vacuum.unit is u.Angstrom
    vacuum = airtovac(air.to(u.nm))
    #
    # Due to numeric funkiness, 2000 -> 1999.999999999 < 2000.
    #
    assert np.allclose(vacuum.value,
                       np.array([180.0, 190.0, 200.0,
                                 210.06664, 220.06868, 230.07083]))
    assert vacuum.unit is u.nm


def test_cirrange():
    ra1 = np.linspace(-4.0*np.pi, 4.0*np.pi, 100)
    ra2 = cirrange(ra1, radians=True)
    assert (ra2 == (ra1 % (2.0 * np.pi))).all()
    ra1 = np.rad2deg(ra1)
    ra2 = cirrange(ra1)
    assert (ra2 == (ra1 % 360.0)).all()


def test_flegendre():
    x = np.array([-1, -0.5, 0, 0.5, 1], dtype='d')
    #
    # Test order
    #
    with pytest.raises(ValueError):
        f = flegendre(x, 0)
    #
    # m = 1
    #
    f = flegendre(x, 1)
    assert (f == np.ones((1, x.size), dtype='d')).all()
    #
    # m = 2
    #
    f = flegendre(x, 2)
    foo = np.ones((2, x.size), dtype='d')
    foo[1, :] = x
    assert np.allclose(f, foo)
    #
    # m = 3
    #
    f = flegendre(x, 3)
    foo = np.ones((3, x.size), dtype='d')
    foo[1, :] = x
    foo[2, :] = 0.5 * (3.0*x**2 - 1.0)
    assert np.allclose(f, foo)
    #
    # m = 4
    #
    f = flegendre(x, 4)
    foo = np.ones((4, x.size), dtype='d')
    foo[1, :] = x
    foo[2, :] = 0.5*(3.0*x**2 - 1.0)
    foo[3, :] = 0.5*(5.0*x**3 - 3.0*x)
    assert np.allclose(f, foo)
    #
    # random float
    #
    f = flegendre(2.88, 3)
    assert np.allclose(f, np.array([[1.00], [2.88], [11.9416]]))


def test_gcirc():
    np.random.seed(137)
    #
    # Start in radians
    #
    offset = 5.0e-6  # approx 1 arcsec
    ra1 = 2.0 * np.pi * np.random.rand(100)
    dec1 = np.pi/2.0 - np.arccos(2.0*np.random.rand(100) - 1.0)
    ra2 = ra1 + offset
    ra2 = np.where((ra2 > 2.0*np.pi), ra2 - 2.0*np.pi, ra2)
    dec2 = np.where((dec1 > 0), dec1 - offset, dec1 + offset)
    deldec2 = (dec2-dec1)/2.0
    delra2 = (ra2-ra1)/2.0
    sindis = np.sqrt(np.sin(deldec2) * np.sin(deldec2) +
                     np.cos(dec1) * np.cos(dec2) *
                     np.sin(delra2) * np.sin(delra2))
    dis = 2.0*np.arcsin(sindis)
    #
    # units = 0
    #
    d0 = gcirc(ra1, dec1, ra2, dec2, units=0)
    assert np.allclose(d0, dis)
    #
    # units = 2
    #
    d0 = gcirc(np.rad2deg(ra1)/15.0, np.rad2deg(dec1),
               np.rad2deg(ra2)/15.0, np.rad2deg(dec2), units=1)
    assert np.allclose(d0, np.rad2deg(dis)*3600.0)
    #
    # units = 2
    #
    d0 = gcirc(np.rad2deg(ra1), np.rad2deg(dec1), np.rad2deg(ra2),
               np.rad2deg(dec2), units=2)
    assert np.allclose(d0, np.rad2deg(dis)*3600.0)
    #
    # Units = whatever
    #
    with pytest.raises(ValueError):
        d0 = gcirc(ra1, dec1, ra2, dec2, units=5)


def test_get_juldate():
    now = get_juldate()
    assert now > 2400000.5
    assert get_juldate(0) == 40587.0 + 2400000.5
    assert get_juldate(86400) == 40588.0 + 2400000.5


def test_vactoair():
    air = vactoair(1900.0)
    assert air == 1900.0
    air = vactoair(2000.0)
    assert np.allclose(air, 1999.3526)
    vacuum = np.array([1800.0, 1900.0, 2000.0, 2100.0, 2200.0, 2300.0])
    air = vactoair(vacuum)
    assert np.allclose(air, np.array([1800.0, 1900.0, 1999.3526,
                                      2099.3337, 2199.3133, 2299.2918]))
    air = vactoair(np.array([1800.0, 1850.0, 1900.0]))
    assert np.allclose(air, np.array([1800.0, 1850.0, 1900.0]))
    #
    # Regression test for #8.
    #
    wave = vacuum.reshape(2, 3)
    air = vactoair(wave)
    assert np.allclose(air,
                       np.array([[1800.0, 1900.0, 1999.3526],
                                 [2099.3337, 2199.3133, 2299.2918]]))
    #
    # Test with units.
    #
    vacuum = vacuum * u.Angstrom
    air = vactoair(vacuum)
    assert np.allclose(air.value,
                       np.array([1800.0, 1900.0, 1999.3526,
                                 2099.3337, 2199.3133, 2299.2918]))
    assert air.unit is u.Angstrom
    air = vactoair(vacuum.to(u.nm))
    #
    # Due to numeric funkiness, 2000 -> 1999.999999999 < 2000.
    #
    assert np.allclose(air.value,
                       np.array([180.0, 190.0, 200.0,
                                 209.93337, 219.93133, 229.92918]))
    assert air.unit is u.nm



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
====
pydl
====

Python replacements for functions that are part of the `IDL®`_ built-in library, or
part of astronomical `IDL®`_ libraries.  The emphasis is on reproducing results of
the astronomical library functions.  Only the bare minimum of `IDL®`_ built-in
functions are implemented to support this.

.. _`IDL®`: https://www.nv5geospatialsoftware.com/Products/IDL
"""
import os
from astropy.tests.runner import TestRunner

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

from .file_lines import file_lines
from .median import median
from .pcomp import pcomp
from .rebin import rebin
from .smooth import smooth
from .uniq import uniq


# Create the test function for self test
test = TestRunner.make_test_runner_in(os.path.dirname(__file__))
test.__test__ = False


class PydlException(Exception):
    """Base class for exceptions raised in PyDL functions.
    """
    pass


__all__ = ['file_lines', 'median', 'pcomp', 'rebin', 'smooth', 'uniq',
           'PydlException']

# Clean up namespace
del os
del TestRunner



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""Test the functions in pydl.pydlutils.rgbcolor.
"""
import pytest
import numpy as np
from .. import PydlutilsUserWarning
from ..rgbcolor import (nw_arcsinh, nw_cut_to_box, nw_float_to_byte,
                        nw_scale_rgb)


def test_nw_arcsinh():
    colors = np.random.random((10, 10))
    with pytest.raises(ValueError):
        fitted_colors = nw_arcsinh(colors)
    colors = np.random.random((10, 10, 5))
    with pytest.raises(ValueError):
        fitted_colors = nw_arcsinh(colors)
    colors = np.random.random((10, 10, 3))
    fitted_colors = nw_arcsinh(colors, nonlinearity=0)
    assert (fitted_colors == colors).all()
    colors = np.ones((2, 2, 3))
    fac = np.arcsinh(9.0)/9.0
    fitted_colors = nw_arcsinh(colors)
    assert np.allclose(fitted_colors, fac)


def test_nw_cut_to_box():
    colors = np.random.random((10, 10))
    with pytest.raises(ValueError):
        boxed_colors = nw_cut_to_box(colors)
    colors = np.random.random((10, 10, 5))
    with pytest.raises(ValueError):
        boxed_colors = nw_cut_to_box(colors)
    colors = np.random.random((10, 10, 3))
    with pytest.raises(ValueError):
        boxed_colors = nw_cut_to_box(colors, origin=(1.0, 1.0))
    boxed_colors = nw_cut_to_box(colors)
    assert np.allclose(boxed_colors, colors)


def test_nw_float_to_byte():
    colors = np.zeros((10, 10, 3), dtype=np.float32)
    byte_colors = nw_float_to_byte(colors)
    assert (byte_colors == 0).all()
    colors = np.ones((10, 10, 3), dtype=np.float32)
    byte_colors = nw_float_to_byte(colors)
    assert (byte_colors == 255).all()
    with pytest.warns(PydlutilsUserWarning) as w:
        byte_colors = nw_float_to_byte(colors, bits=16)
    assert len(w) > 0


def test_nw_scale_rgb():
    colors = np.random.random((10, 10))
    with pytest.raises(ValueError):
        scaled_colors = nw_scale_rgb(colors)
    colors = np.random.random((10, 10, 5))
    with pytest.raises(ValueError):
        scaled_colors = nw_scale_rgb(colors)
    colors = np.random.random((10, 10, 3))
    with pytest.raises(ValueError):
        scaled_colors = nw_scale_rgb(colors, scales=(1.0, 1.0))
    colors = np.ones((2, 2, 3))
    scaled_colors = nw_scale_rgb(colors, scales=(2.0, 2.0, 2.0))
    assert np.allclose(scaled_colors, 2.0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
This subpackage implements functions from the photoop package.
"""
#
# Define this early on so that submodules can use it
#
from .. import PydlException


class PhotoopException(PydlException):
    """Exceptions raised by :mod:`pydl.photoop` that don't fit into a
    standard exception class like :exc:`ValueError`.
    """
    pass


__all__ = ['PhotoopException']
